@@ -21,6 +21,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
+#include <sys/xattr.h>
 #include <time.h>
 #include <unistd.h>
 
@@ -85,6 +86,13 @@ static struct sql_table {
 	    "rdev INTEGER",
 	  NULL,
 	  NULL },
+	{ "xattrs",
+	  "id INTEGER PRIMARY KEY, "
+	    "file_id INTEGER REFERENCES files(id) ON DELETE CASCADE, "
+	    "name VARCHAR, "
+	    "value VARCHAR",
+	  NULL,
+	  NULL },
 	{ NULL, NULL, NULL, NULL },
 }, log_tables[] = {
 	{ "logs",
@@ -114,6 +122,7 @@ static struct sql_index {
 /*	{ "files__path", "files", "path" }, */
 	{ "files__path_dev_ino", "files", "path, dev, ino" },
 	{ "files__dev_ino", "files", "dev, ino" },
+	{ "xattrs__file", "xattrs", "file_id" },
 	{ NULL, NULL, NULL },
 }, log_indexes[] = {
 	{ NULL, NULL, NULL },
@@ -136,6 +145,7 @@ static char *file_pragmas[] = {
 	 * need.
 	 */
 	"PRAGMA synchronous = OFF;",
+	"PRAGMA foreign_keys = ON;",
 	NULL
 };
 
@@ -365,6 +375,7 @@ make_tables(sqlite3 *db,
 
 	for (i = 0; sql_tables[i].name; ++i) {
 		found = 0;
+		printf("considering table %s\n", sql_tables[i].name);
 		for (j = 1; j <= rows; ++j) {
 			if (!strcmp(existing[j], sql_tables[i].name)) {
 				found = 1;
@@ -441,14 +452,14 @@ make_tables(sqlite3 *db,
 			dberr(db, "not done after the single row we expected?", rc);
 			return 1;
 		}
-		pseudo_debug(2, "existing database version: %d\n", version);
+		pseudo_debug(PDBGF_DB, "existing database version: %d\n", version);
 		rc = sqlite3_finalize(stmt);
 		if (rc) {
 			dberr(db, "couldn't finalize version check");
 			return 1;
 		}
 	} else {
-		pseudo_debug(2, "no existing database version\n");
+		pseudo_debug(PDBGF_DB, "no existing database version\n");
 		version = -1;
 	}
 	for (m = sql_migrations; m->sql; ++m)
@@ -462,10 +473,10 @@ make_tables(sqlite3 *db,
 		version = available_migrations - 1;
 	for (m = sql_migrations + (version + 1); m->sql; ++m) {
 		int migration = (m - sql_migrations);
-		pseudo_debug(3, "considering migration %d\n", migration);
+		pseudo_debug(PDBGF_DB, "considering migration %d\n", migration);
 		if (version >= migration)
 			continue;
-		pseudo_debug(2, "running migration %d\n", migration);
+		pseudo_debug(PDBGF_DB, "running migration %d\n", migration);
 		rc = sqlite3_prepare_v2(db,
 				m->sql,
 				strlen(m->sql),
@@ -512,7 +523,7 @@ make_tables(sqlite3 *db,
 			sqlite3_finalize(update_version);
 			return 1;
 		} else {
-			pseudo_debug(3, "update of migrations (after %d) fine.\n",
+			pseudo_debug(PDBGF_DB, "update of migrations (after %d) fine.\n",
 				migration);
 		}
 		sqlite3_finalize(update_version);
@@ -526,7 +537,7 @@ make_tables(sqlite3 *db,
 /* registered with atexit */
 static void
 cleanup_db(void) {
-	pseudo_debug(1, "server exiting\n");
+	pseudo_debug(PDBGF_SERVER, "server exiting\n");
 #ifdef USE_MEMORY_DB
         if (real_file_db) {
                 pdb_backup();
@@ -587,6 +598,8 @@ get_db(struct database_info *dbinfo) {
 	if (dbinfo->pragmas) {
 		for (i = 0; dbinfo->pragmas[i]; ++i) {
 			rc = sqlite3_exec(db, dbinfo->pragmas[i], NULL, NULL, &errmsg);
+			pseudo_debug(PDBGF_SQL | PDBGF_VERBOSE, "executed pragma: '%s', rc %d.\n",
+				dbinfo->pragmas[i], rc);
 			if (rc) {
 				dberr(db, dbinfo->pragmas[i]);
 			}
@@ -1124,7 +1137,7 @@ pdb_query(char *stmt_type, pseudo_query_t *traits, unsigned long fields, int uni
 	}
 	if (want_results)
 		frag(sql, "ORDER BY %s %s;", order_by, order_dir);
-	pseudo_debug(1, "created SQL: <%s>\n", sql->data);
+	pseudo_debug(PDBGF_SQL, "created SQL: <%s>\n", sql->data);
 
 	/* second, prepare it */
 	rc = sqlite3_prepare_v2(log_db, sql->data, strlen(sql->data), &stmt, NULL);
@@ -1356,7 +1369,7 @@ log_entry_free(log_entry *e) {
  * or 'NAMELESS FILE'.
  */
 int
-pdb_link_file(pseudo_msg_t *msg) {
+pdb_link_file(pseudo_msg_t *msg, long long *row) {
 	static sqlite3_stmt *insert;
 	int rc;
 	char *sql = "INSERT INTO files "
@@ -1388,7 +1401,7 @@ pdb_link_file(pseudo_msg_t *msg) {
 	sqlite3_bind_int(insert, 5, msg->gid);
 	sqlite3_bind_int(insert, 6, msg->mode);
 	sqlite3_bind_int(insert, 7, msg->rdev);
-	pseudo_debug(2, "linking %s: dev %llu, ino %llu, mode %o, owner %d\n",
+	pseudo_debug(PDBGF_DB, "linking %s: dev %llu, ino %llu, mode %o, owner %d\n",
 		(msg->pathlen ? msg->path : "<nil> (as NAMELESS FILE)"),
 		(unsigned long long) msg->dev, (unsigned long long) msg->ino,
 		(int) msg->mode, msg->uid);
@@ -1397,6 +1410,10 @@ pdb_link_file(pseudo_msg_t *msg) {
 	if (rc != SQLITE_DONE) {
 		dberr(file_db, "insert may have failed (rc %d)", rc);
 	}
+	/* some users care what the row ID is */
+	if (row) {
+		*row = sqlite3_last_insert_rowid(file_db);
+	}
 	sqlite3_reset(insert);
 	sqlite3_clear_bindings(insert);
 	return rc != SQLITE_DONE;
@@ -1455,7 +1472,7 @@ pdb_update_file_path(pseudo_msg_t *msg) {
 		}
 	}
 	if (!msg || !msg->pathlen) {
-		pseudo_debug(1, "can't update a file without a message or path.\n");
+		pseudo_debug(PDBGF_DB, "can't update a file without a message or path.\n");
 		return 1;
 	}
 	sqlite3_bind_text(update, 1, msg->path, -1, SQLITE_STATIC);
@@ -1496,7 +1513,7 @@ pdb_may_unlink_file(pseudo_msg_t *msg, int deleting) {
 		sqlite3_bind_int(mark_file, 1, deleting);
 		sqlite3_bind_text(mark_file, 2, msg->path, -1, SQLITE_STATIC);
 	} else {
-		pseudo_debug(1, "cannot mark a file for pending deletion without a path.");
+		pseudo_debug(PDBGF_DB, "cannot mark a file for pending deletion without a path.");
 		return 1;
 	}
 	file_db_dirty = 1;
@@ -1506,7 +1523,7 @@ pdb_may_unlink_file(pseudo_msg_t *msg, int deleting) {
 		return 1;
 	}
 	exact = sqlite3_changes(file_db);
-	pseudo_debug(3, "(exact %d) ", exact);
+	pseudo_debug(PDBGF_DB, "(exact %d) ", exact);
 	sqlite3_reset(mark_file);
 	sqlite3_clear_bindings(mark_file);
 	/* indicate whether we marked something */
@@ -1540,7 +1557,7 @@ pdb_cancel_unlink_file(pseudo_msg_t *msg) {
 	if (msg->pathlen) {
 		sqlite3_bind_text(mark_file, 1, msg->path, -1, SQLITE_STATIC);
 	} else {
-		pseudo_debug(1, "cannot unmark a file for pending deletion without a path.");
+		pseudo_debug(PDBGF_DB, "cannot unmark a file for pending deletion without a path.");
 		return 1;
 	}
 	file_db_dirty = 1;
@@ -1549,7 +1566,7 @@ pdb_cancel_unlink_file(pseudo_msg_t *msg) {
 		dberr(file_db, "unmark for deletion may have failed");
 	}
 	exact = sqlite3_changes(file_db);
-	pseudo_debug(3, "(exact %d) ", exact);
+	pseudo_debug(PDBGF_DB, "(exact %d) ", exact);
 	sqlite3_reset(mark_file);
 	sqlite3_clear_bindings(mark_file);
 	return rc != SQLITE_DONE;
@@ -1586,7 +1603,7 @@ pdb_did_unlink_files(int deleting) {
 		dberr(file_db, "cleanup of files marked for deletion may have failed");
 	}
 	exact = sqlite3_changes(file_db);
-	pseudo_debug(3, "(exact %d)\n", exact);
+	pseudo_debug(PDBGF_DB, "(exact %d)\n", exact);
 	sqlite3_reset(delete_exact);
 	sqlite3_clear_bindings(delete_exact);
 	return rc != SQLITE_DONE;
@@ -1611,7 +1628,7 @@ pdb_did_unlink_file(char *path, int deleting) {
 		}
 	}
 	if (!path) {
-		pseudo_debug(1, "cannot unlink a file without a path.");
+		pseudo_debug(PDBGF_DB, "cannot unlink a file without a path.");
 		return 1;
 	}
 	sqlite3_bind_text(delete_exact, 1, path, -1, SQLITE_STATIC);
@@ -1622,7 +1639,7 @@ pdb_did_unlink_file(char *path, int deleting) {
 		dberr(file_db, "cleanup of file marked for deletion may have failed");
 	}
 	exact = sqlite3_changes(file_db);
-	pseudo_debug(3, "(exact %d)\n", exact);
+	pseudo_debug(PDBGF_DB, "(exact %d)\n", exact);
 	sqlite3_reset(delete_exact);
 	sqlite3_clear_bindings(delete_exact);
 	return rc != SQLITE_DONE;
@@ -1652,7 +1669,7 @@ pdb_unlink_file(pseudo_msg_t *msg) {
 	if (msg->pathlen) {
 		sqlite3_bind_text(delete_exact, 1, msg->path, -1, SQLITE_STATIC);
 	} else {
-		pseudo_debug(1, "cannot unlink a file without a path.");
+		pseudo_debug(PDBGF_DB, "cannot unlink a file without a path.");
 		return 1;
 	}
 	file_db_dirty = 1;
@@ -1661,7 +1678,7 @@ pdb_unlink_file(pseudo_msg_t *msg) {
 		dberr(file_db, "delete exact by path may have failed");
 	}
 	exact = sqlite3_changes(file_db);
-	pseudo_debug(3, "(exact %d) ", exact);
+	pseudo_debug(PDBGF_DB, "(exact %d) ", exact);
 	sqlite3_reset(delete_exact);
 	sqlite3_clear_bindings(delete_exact);
 	return rc != SQLITE_DONE;
@@ -1700,7 +1717,7 @@ pdb_unlink_contents(pseudo_msg_t *msg) {
 		sqlite3_bind_text(delete_sub, 1, msg->path, -1, SQLITE_STATIC);
 		sqlite3_bind_text(delete_sub, 2, msg->path, -1, SQLITE_STATIC);
 	} else {
-		pseudo_debug(1, "cannot unlink a file without a path.");
+		pseudo_debug(PDBGF_DB, "cannot unlink a file without a path.");
 		return 1;
 	}
 	file_db_dirty = 1;
@@ -1709,7 +1726,7 @@ pdb_unlink_contents(pseudo_msg_t *msg) {
 		dberr(file_db, "delete sub by path may have failed");
 	}
 	sub = sqlite3_changes(file_db);
-	pseudo_debug(3, "(sub %d) ", sub);
+	pseudo_debug(PDBGF_DB, "(sub %d) ", sub);
 	sqlite3_reset(delete_sub);
 	sqlite3_clear_bindings(delete_sub);
 	return rc != SQLITE_DONE;
@@ -1754,14 +1771,14 @@ pdb_rename_file(const char *oldpath, pseudo_msg_t *msg) {
 		return 1;
 	}
 	if (!msg->pathlen) {
-		pseudo_debug(1, "rename: No path provided (ino %llu)\n", (unsigned long long) msg->ino);
+		pseudo_debug(PDBGF_DB, "rename: No path provided (ino %llu)\n", (unsigned long long) msg->ino);
 		return 1;
 	}
 	if (!oldpath) {
-		pseudo_debug(1, "rename: No old path for %s\n", msg->path);
+		pseudo_debug(PDBGF_DB, "rename: No old path for %s\n", msg->path);
 		return 1;
 	}
-	pseudo_debug(2, "rename: Changing %s to %s\n", oldpath, msg->path);
+	pseudo_debug(PDBGF_DB, "rename: Changing %s to %s\n", oldpath, msg->path);
 	rc = sqlite3_bind_text(update_exact, 1, msg->path, -1, SQLITE_STATIC);
 	rc = sqlite3_bind_text(update_exact, 2, oldpath, -1, SQLITE_STATIC);
 	rc = sqlite3_bind_text(update_sub, 1, oldpath, -1, SQLITE_STATIC);
@@ -1829,7 +1846,7 @@ pdb_renumber_all(dev_t from, dev_t to) {
 	}
 	sqlite3_reset(update);
 	sqlite3_clear_bindings(update);
-	pseudo_debug(2, "updating device dev %llu to %llu\n",
+	pseudo_debug(PDBGF_DB, "updating device dev %llu to %llu\n",
 		(unsigned long long) from, (unsigned long long) to);
 	return rc != SQLITE_DONE;
 }
@@ -1880,7 +1897,7 @@ pdb_update_inode(pseudo_msg_t *msg) {
 	}
 	sqlite3_reset(update);
 	sqlite3_clear_bindings(update);
-	pseudo_debug(2, "updating path %s to dev %llu, ino %llu\n",
+	pseudo_debug(PDBGF_DB, "updating path %s to dev %llu, ino %llu\n",
 		msg->path,
 		(unsigned long long) msg->dev, (unsigned long long) msg->ino);
 	return rc != SQLITE_DONE;
@@ -1925,7 +1942,7 @@ pdb_update_file(pseudo_msg_t *msg) {
 	}
 	sqlite3_reset(update);
 	sqlite3_clear_bindings(update);
-	pseudo_debug(2, "updating dev %llu, ino %llu, new mode %o, owner %d\n",
+	pseudo_debug(PDBGF_DB, "updating dev %llu, ino %llu, new mode %o, owner %d\n",
 		(unsigned long long) msg->dev, (unsigned long long) msg->ino,
 		(int) msg->mode, msg->uid);
 	return rc != SQLITE_DONE;
@@ -1933,7 +1950,7 @@ pdb_update_file(pseudo_msg_t *msg) {
 
 /* find file using both path AND dev/inode as key */
 int
-pdb_find_file_exact(pseudo_msg_t *msg) {
+pdb_find_file_exact(pseudo_msg_t *msg, long long *row) {
 	static sqlite3_stmt *select;
 	int rc;
 	char *sql = "SELECT * FROM files WHERE path = ? AND dev = ? AND ino = ?;";
@@ -1961,6 +1978,9 @@ pdb_find_file_exact(pseudo_msg_t *msg) {
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
+		if (row) {
+			*row = sqlite3_column_int64(select, 0);
+		}
 		msg->uid = (unsigned long) sqlite3_column_int64(select, 4);
 		msg->gid = (unsigned long) sqlite3_column_int64(select, 5);
 		msg->mode = (unsigned long) sqlite3_column_int64(select, 6);
@@ -1969,7 +1989,7 @@ pdb_find_file_exact(pseudo_msg_t *msg) {
 		rc = 0;
 		break;
 	case SQLITE_DONE:
-		pseudo_debug(3, "find_exact: sqlite_done on first row\n");
+		pseudo_debug(PDBGF_DB, "find_exact: sqlite_done on first row\n");
 		rc = 1;
 		break;
 	default:
@@ -1984,7 +2004,7 @@ pdb_find_file_exact(pseudo_msg_t *msg) {
 
 /* find file using path as a key */
 int
-pdb_find_file_path(pseudo_msg_t *msg) {
+pdb_find_file_path(pseudo_msg_t *msg, long long *row) {
 	static sqlite3_stmt *select;
 	int rc;
 	char *sql = "SELECT * FROM files WHERE path = ?;";
@@ -2015,6 +2035,9 @@ pdb_find_file_path(pseudo_msg_t *msg) {
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
+		if (row) {
+			*row = sqlite3_column_int64(select, 0);
+		}
 		msg->dev = sqlite3_column_int64(select, 2);
 		msg->ino = sqlite3_column_int64(select, 3);
 		msg->uid = sqlite3_column_int64(select, 4);
@@ -2025,7 +2048,7 @@ pdb_find_file_path(pseudo_msg_t *msg) {
 		rc = 0;
 		break;
 	case SQLITE_DONE:
-		pseudo_debug(3, "find_path: sqlite_done on first row\n");
+		pseudo_debug(PDBGF_DB, "find_path: sqlite_done on first row\n");
 		rc = 1;
 		break;
 	default:
@@ -2075,7 +2098,7 @@ pdb_get_file_path(pseudo_msg_t *msg) {
 		}
 		break;
 	case SQLITE_DONE:
-		pseudo_debug(3, "find_dev: sqlite_done on first row\n");
+		pseudo_debug(PDBGF_DB, "find_dev: sqlite_done on first row\n");
 		response = 0;
 		break;
 	default:
@@ -2090,7 +2113,7 @@ pdb_get_file_path(pseudo_msg_t *msg) {
 
 /* find file using dev/inode as key */
 int
-pdb_find_file_dev(pseudo_msg_t *msg) {
+pdb_find_file_dev(pseudo_msg_t *msg, long long *row) {
 	static sqlite3_stmt *select;
 	int rc;
 	char *sql = "SELECT * FROM files WHERE dev = ? AND ino = ?;";
@@ -2114,6 +2137,9 @@ pdb_find_file_dev(pseudo_msg_t *msg) {
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
+		if (row) {
+			*row = sqlite3_column_int64(select, 0);
+		}
 		msg->uid = (unsigned long) sqlite3_column_int64(select, 4);
 		msg->gid = (unsigned long) sqlite3_column_int64(select, 5);
 		msg->mode = (unsigned long) sqlite3_column_int64(select, 6);
@@ -2122,7 +2148,7 @@ pdb_find_file_dev(pseudo_msg_t *msg) {
 		rc = 0;
 		break;
 	case SQLITE_DONE:
-		pseudo_debug(3, "find_dev: sqlite_done on first row\n");
+		pseudo_debug(PDBGF_DB, "find_dev: sqlite_done on first row\n");
 		rc = 1;
 		break;
 	default:
@@ -2135,11 +2161,286 @@ pdb_find_file_dev(pseudo_msg_t *msg) {
 	return rc;
 }
 
+int
+pdb_get_xattr(long long file_id, char **value, size_t *len) {
+	static sqlite3_stmt *select;
+	int rc;
+	char *response;
+	size_t length;
+	char *sql = "SELECT value FROM xattrs WHERE file_id = ? AND name = ?;";
+
+	if (!file_db && get_dbs()) {
+		pseudo_diag("%s: database error.\n", __func__);
+		return 0;
+	}
+	if (!select) {
+		rc = sqlite3_prepare_v2(file_db, sql, strlen(sql), &select, NULL);
+		if (rc) {
+			dberr(file_db, "couldn't prepare SELECT statement");
+			return 1;
+		}
+	}
+	pseudo_debug(PDBGF_XATTR, "requested xattr named '%s' for file %lld\n", *value, file_id);
+	sqlite3_bind_int(select, 1, file_id);
+	rc = sqlite3_bind_text(select, 2, *value, -1, SQLITE_STATIC);
+	if (rc) {
+		dberr(file_db, "couldn't bind xattr name to SELECT.");
+		return 1;
+	}
+	rc = sqlite3_step(select);
+	switch (rc) {
+	case SQLITE_ROW:
+		response = (char *) sqlite3_column_text(select, 0);
+		length = sqlite3_column_bytes(select, 0);
+		pseudo_debug(PDBGF_XATTR, "got %d-byte results: '%s'\n",
+			(int) length, response);
+		if (response && length >= 1) {
+			/* not a strdup because the values can contain
+			 * arbitrary bytes.
+			 */
+			*value = malloc(length);
+			memcpy(*value, response, length);
+			*len = length;
+			rc = 0;
+		} else {
+			*value = NULL;
+			*len = 0;
+			rc = 1;
+		}
+		break;
+	case SQLITE_DONE:
+		pseudo_debug(PDBGF_DB, "find_exact: sqlite_done on first row\n");
+		rc = 1;
+		break;
+	default:
+		dberr(file_db, "find_exact: select returned neither a row nor done");
+		rc = 1;
+		break;
+	}
+	sqlite3_reset(select);
+	sqlite3_clear_bindings(select);
+	return rc;
+}
+
+int
+pdb_list_xattr(long long file_id, char **value, size_t *len) {
+	static sqlite3_stmt *select;
+	size_t allocated = 0;
+	size_t used = 0;
+	char *buffer = 0;
+	int rc;
+	char *sql = "SELECT name FROM xattrs WHERE file_id = ?;";
+
+	/* if we don't have a record of the file, it must not have
+	 * any extended attributes...
+	 */
+	if (file_id == -1) {
+		*value = NULL;
+		*len = 0;
+		return 0;
+	}
+
+	if (!file_db && get_dbs()) {
+		pseudo_diag("%s: database error.\n", __func__);
+		return 0;
+	}
+	if (!select) {
+		rc = sqlite3_prepare_v2(file_db, sql, strlen(sql), &select, NULL);
+		if (rc) {
+			dberr(file_db, "couldn't prepare SELECT statement");
+			return 1;
+		}
+	}
+	sqlite3_bind_int(select, 1, file_id);
+	do {
+		rc = sqlite3_step(select);
+		if (rc == SQLITE_ROW) {
+			char *value = (char *) sqlite3_column_text(select, 0);
+			size_t len = sqlite3_column_bytes(select, 0);
+			if (!buffer) {
+				allocated = round_up(len, 256);
+				buffer = malloc(allocated);
+			}
+			if (used + len + 2 > allocated) {
+				size_t new_allocated = round_up(used + len + 2, 256);
+				char *new_buffer = malloc(new_allocated);
+				memcpy(new_buffer, buffer, used);
+				free(buffer);
+				allocated = new_allocated;
+				buffer = new_buffer;
+			}
+			memcpy(buffer + used, value, len);
+			buffer[used + len] = '\0';
+			used = used + len + 1;
+		} else if (rc == SQLITE_DONE) {
+			*value = buffer;
+			*len = used;
+		} else {
+			dberr(file_db, "non-row response from select?");
+			free(buffer);
+			*value = NULL;
+			*len = 0;
+		}
+	} while (rc == SQLITE_ROW);
+	sqlite3_reset(select);
+	sqlite3_clear_bindings(select);
+	return rc != SQLITE_DONE;
+}
+
+int
+pdb_remove_xattr(long long file_id, char *value, size_t len) {
+	static sqlite3_stmt *delete;
+	int rc;
+	char *sql = "DELETE FROM xattrs WHERE file_id = ? AND name = ?;";
+
+	if (!file_db && get_dbs()) {
+		pseudo_diag("%s: database error.\n", __func__);
+		return 0;
+	}
+	if (!delete) {
+		rc = sqlite3_prepare_v2(file_db, sql, strlen(sql), &delete, NULL);
+		if (rc) {
+			dberr(file_db, "couldn't prepare DELETE statement");
+			return 1;
+		}
+	}
+	sqlite3_bind_int(delete, 1, file_id);
+	rc = sqlite3_bind_text(delete, 2, value, len, SQLITE_STATIC);
+	if (rc) {
+		dberr(file_db, "couldn't bind xattr name to DELETE.");
+		return 1;
+	}
+	file_db_dirty = 1;
+	rc = sqlite3_step(delete);
+	if (rc != SQLITE_DONE) {
+		dberr(file_db, "delete xattr may have failed");
+	}
+	sqlite3_reset(delete);
+	sqlite3_clear_bindings(delete);
+	return rc != SQLITE_DONE;
+}
+
+int
+pdb_set_xattr(long long file_id, char *value, size_t len, int flags) {
+	static sqlite3_stmt *select, *update, *insert;
+	int rc;
+	long long existing_row = -1;
+	char *select_sql = "SELECT id FROM xattrs WHERE file_id = ? AND name = ?;";
+	char *insert_sql = "INSERT INTO xattrs "
+		    " ( file_id, name, value ) "
+		    " VALUES (?, ?, ?);";
+	char *update_sql = "UPDATE xattrs SET value = ? WHERE id = ?;";
+	char *vname = value;
+	size_t vlen;
+
+	if (!file_db && get_dbs()) {
+		pseudo_diag("%s: database error.\n", __func__);
+		return 0;
+	}
+	if (!select) {
+		rc = sqlite3_prepare_v2(file_db, select_sql, strlen(select_sql), &select, NULL);
+		if (rc) {
+			dberr(file_db, "couldn't prepare SELECT statement");
+			return 1;
+		}
+	}
+	sqlite3_bind_int(select, 1, file_id);
+	rc = sqlite3_bind_text(select, 2, value, -1, SQLITE_STATIC);
+	if (rc) {
+		dberr(file_db, "couldn't bind xattr name to SELECT.");
+		return 1;
+	}
+	rc = sqlite3_step(select);
+	switch (rc) {
+	case SQLITE_ROW:
+		existing_row = sqlite3_column_int64(select, 0);
+		break;
+	case SQLITE_DONE:
+		pseudo_debug(PDBGF_DB | PDBGF_VERBOSE, "find_exact: sqlite_done on first row\n");
+		existing_row = -1;
+		break;
+	default:
+		dberr(file_db, "set_xattr: select returned neither a row nor done");
+		rc = 1;
+		break;
+	}
+	sqlite3_reset(select);
+	sqlite3_clear_bindings(select);
+	if (flags == XATTR_CREATE && existing_row != -1) {
+		pseudo_debug(PDBGF_DB, "XATTR_CREATE with an existing row, failing.");
+		return 1;
+	}
+	if (flags == XATTR_REPLACE && existing_row == -1) {
+		pseudo_debug(PDBGF_DB, "XATTR_REPLACE with no existing row, failing.");
+		return 1;
+	}
+	/* the material after the name is the value */
+	vlen = strlen(value);
+	len = len - (vlen + 1);
+	value = value + vlen + 1;
+	pseudo_debug(PDBGF_XATTR, "trying to set a value for %lld: name is '%s' [%d/%d bytes], value is '%s' [%d bytes]. Existing row %lld.\n",
+		file_id, vname, (int) vlen, (int) (len + vlen + 1), value, (int) len, existing_row);
+	if (existing_row != -1) {
+		/* update */
+		if (!update) {
+			rc = sqlite3_prepare_v2(file_db, update_sql, strlen(update_sql), &update, NULL);
+			if (rc) {
+				dberr(file_db, "couldn't prepare UPDATE statement");
+				return 1;
+			}
+		}
+		rc = sqlite3_bind_text(update, 1, value, len, SQLITE_STATIC);
+		if (rc) {
+			dberr(file_db, "couldn't bind xattr value to UPDATE.");
+			return 1;
+		}
+		sqlite3_bind_int(update, 2, existing_row);
+		file_db_dirty = 1;
+		rc = sqlite3_step(update);
+		if (rc != SQLITE_DONE) {
+			dberr(file_db, "update xattr may have failed");
+		}
+		sqlite3_reset(update);
+		sqlite3_clear_bindings(update);
+		return rc != SQLITE_DONE;
+	} else {
+		/* insert */
+		if (!insert) {
+			rc = sqlite3_prepare_v2(file_db, insert_sql, strlen(insert_sql), &insert, NULL);
+			if (rc) {
+				dberr(file_db, "couldn't prepare INSERT statement");
+				return 1;
+			}
+		}
+		pseudo_debug(PDBGF_XATTR, "insert should be getting ID %lld\n", file_id);
+		sqlite3_bind_int64(insert, 1, file_id);
+		rc = sqlite3_bind_text(insert, 2, vname, -1, SQLITE_STATIC);
+		if (rc) {
+			dberr(file_db, "couldn't bind xattr name to INSERT statement");
+			return 1;
+		}
+		rc = sqlite3_bind_text(insert, 3, value, len, SQLITE_STATIC);
+		if (rc) {
+			dberr(file_db, "couldn't bind xattr value to INSERT statement");
+			return 1;
+		}
+		file_db_dirty = 1;
+		rc = sqlite3_step(insert);
+		if (rc != SQLITE_DONE) {
+			dberr(file_db, "insert xattr may have failed");
+		}
+		sqlite3_reset(insert);
+		sqlite3_clear_bindings(insert);
+		return rc != SQLITE_DONE;
+	}
+	return rc;
+}
+
 /* find file using only inode as key.  Unused for now, planned to come
  * in for NFS usage.
  */
 int
-pdb_find_file_ino(pseudo_msg_t *msg) {
+pdb_find_file_ino(pseudo_msg_t *msg, long long *row) {
 	static sqlite3_stmt *select;
 	int rc;
 	char *sql = "SELECT * FROM files WHERE ino = ?;";
@@ -2162,6 +2463,9 @@ pdb_find_file_ino(pseudo_msg_t *msg) {
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
+		if (row) {
+			*row = sqlite3_column_int64(select, 0);
+		}
 		msg->dev = (unsigned long) sqlite3_column_int64(select, 2);
 		msg->uid = (unsigned long) sqlite3_column_int64(select, 4);
 		msg->gid = (unsigned long) sqlite3_column_int64(select, 5);
@@ -2171,7 +2475,7 @@ pdb_find_file_ino(pseudo_msg_t *msg) {
 		rc = 0;
 		break;
 	case SQLITE_DONE:
-		pseudo_debug(3, "find_ino: sqlite_done on first row\n");
+		pseudo_debug(PDBGF_DB, "find_ino: sqlite_done on first row\n");
 		rc = 1;
 		break;
 	default:
@@ -2231,7 +2535,7 @@ pdb_file(pdb_file_list l) {
 		pseudo_diag("couldn't allocate file message.\n");
 		return NULL;
 	}
-	pseudo_debug(2, "pdb_file: '%s'\n", s ? (const char *) s : "<nil>");
+	pseudo_debug(PDBGF_DB, "pdb_file: '%s'\n", s ? (const char *) s : "<nil>");
 	m->dev = sqlite3_column_int64(l->stmt, column++);
 	m->ino = sqlite3_column_int64(l->stmt, column++);
 	m->uid = sqlite3_column_int64(l->stmt, column++);
